#!/usr/bin/env python3

# Challenge sviluppata da Leonardo Tamiano durante il lavoro per la tesi magistrale.

import binascii
import socketserver
import signal
import time
from hashlib import md5
from base64 import b64decode
from base64 import b64encode

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

from secret import FLAG

CHALLENGE_NAME = "Yet Another Oracle"
PORT = 4444

IV = None
KEY = None
CIPHER = None
ENCRYPTED_FLAG = None

def oracle(payload):
    global IV, KEY, CIPHER
    
    try:
        payload = b64decode(payload)
        CIPHER = AES.new(KEY, AES.MODE_CBC, IV)        
        decrypted_payload = CIPHER.decrypt(payload)
        unpadded_payload = unpad(decrypted_payload, AES.block_size)
        return True
    except ValueError as e:
        return False

def challenge(req):
    global IV, KEY, CIPHER, ENCRYPTED_FLAG

    # send initial message
    req.sendall(b'Hi, I\'ve been told to show you this.\n' +\
                b'====================================\n\n' +\
                b'ENCRYPTED_FLAG WITH CBC-AES: ' +\
                ENCRYPTED_FLAG + b'\n\n> ')
    time.sleep(0.2)

    # interact with client's queries and expose oracle.
    #
    # NOTE: this introduces a CBC padding oracle vulnerability.
    while True:
        try:
            client_payload = req.recv(4096)
            if len(client_payload) > 0:
                oracle_output = oracle(client_payload)
                if oracle_output == True:
                    req.sendall(b'OK!\n> ')
                else:
                    req.sendall(b'NOPE\n> ')
            
        except Exception as e:
            print(e)
            exit()

class incoming(socketserver.BaseRequestHandler):
	def handle(self):
		while True:
			challenge(self.request)

class ReusableTCPServer(socketserver.ForkingMixIn, socketserver.TCPServer):
	pass

def challenge_3_main():
    global CHALLENGE_NAME, PORT, IV, KEY, CIPHER, ENCRYPTED_FLAG
    
    IV = get_random_bytes(AES.block_size)
    KEY = get_random_bytes(AES.block_size)
    CIPHER = AES.new(KEY, AES.MODE_CBC, IV)

    # NOTE: Since the attacker is not able to decrypt the first blog,
    # here we put some meaningless values before the flag.
    data_to_encrypt = b"A" * 16 + FLAG
    ENCRYPTED_FLAG = b64encode(CIPHER.encrypt(pad(data_to_encrypt, AES.block_size)))
    
    print(f"[INFO] - Start of challenge: {CHALLENGE_NAME}")
    print(f"[INFO] - Listening on {PORT}...")

    socketserver.TCPServer.allow_reuse_address = True
    server = ReusableTCPServer(("0.0.0.0", PORT), incoming)
    server.serve_forever()

if __name__ == "__main__":
    challenge_3_main()
